package dev.fitko.fitconnect.client.cmd;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.BatchImportException;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.cmd.batch.BatchImporter;
import dev.fitko.fitconnect.client.cmd.batch.CsvImporter;
import dev.fitko.fitconnect.client.cmd.batch.ImportRecord;
import dev.fitko.fitconnect.client.cmd.commands.ListAllSubmissionsCommand;
import dev.fitko.fitconnect.client.cmd.commands.ListOneSubmissionCommand;
import dev.fitko.fitconnect.client.cmd.commands.SendBatchCommand;
import dev.fitko.fitconnect.client.cmd.commands.SendSubmissionCommand;
import dev.fitko.fitconnect.client.cmd.util.AttachmentDataType;
import dev.fitko.fitconnect.client.model.ReceivedSubmission;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class CommandLineClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(CommandLineClient.class);
    public static final String IMPORTED_RECORD_TEMPLATE = "\n\n================== Imported record {} ==================\n";

    private final SendSubmissionCommand sendSubmissionCommand;
    private final ListAllSubmissionsCommand listAllSubmissionsCommand;
    private final ListOneSubmissionCommand listOneSubmissionCommand;
    private final SendBatchCommand sendBatchCommand;

    private final BatchImporter batchImporter;

    private final JCommander jc;

    private final SenderClient.Start senderClient;
    private final SubscriberClient.RequestSubmission subscriberClient;

    public CommandLineClient(final SenderClient.Start senderClient,
                             final SubscriberClient.RequestSubmission subscriberClient) {

        this.sendSubmissionCommand = new SendSubmissionCommand();
        this.listAllSubmissionsCommand = new ListAllSubmissionsCommand();
        this.listOneSubmissionCommand = new ListOneSubmissionCommand();
        this.sendBatchCommand = new SendBatchCommand();

        this.batchImporter = new CsvImporter();

        this.senderClient = senderClient;
        this.subscriberClient = subscriberClient;
        jc = getJCommander();
    }

    public void run(final String... args) {
        final var parsedCommand = parseCommandFromArguments(args);
        if (parsedCommand == null) {
            jc.usage();
            return;
        }
        try {
            executeCommand(parsedCommand);
        } catch (final Exception e) {
            LOGGER.error(e.getMessage());
        }
    }

    private JCommander getJCommander() {
        return JCommander.newBuilder()
                .addCommand(sendSubmissionCommand)
                .addCommand(listAllSubmissionsCommand)
                .addCommand(listOneSubmissionCommand)
                .addCommand(sendBatchCommand)
                .build();
    }

    private void executeCommand(final String parsedCommand) throws IOException {
        switch (parsedCommand) {
            case SendSubmissionCommand.SEND_CMD_NAME:
                sendSubmission(sendSubmissionCommand);
                break;
            case ListAllSubmissionsCommand.LIST_CMD_NAME:
                listSubmissions(listAllSubmissionsCommand);
                break;
            case ListOneSubmissionCommand.GET_CMD_NAME:
                listOneSubmission(listOneSubmissionCommand);
                break;
            case SendBatchCommand.SEND_BATCH_CMD_NAME:
                sendBatch(sendBatchCommand);
                break;
            default:
                break;
        }
    }

    private String parseCommandFromArguments(final String[] args) {
        try {
            jc.parse(args);
            return jc.getParsedCommand();
        } catch (final ParameterException e) {
            LOGGER.error(e.getMessage(), "\n");
            return null;
        }
    }

    private void listOneSubmission(final ListOneSubmissionCommand listOneSubmissionCommand) throws IOException {
        LOGGER.info("Listing submission for id {}", listOneSubmissionCommand.submissionId);
        final var submission = subscriberClient.requestSubmission(listOneSubmissionCommand.submissionId);
        if (submission.isEmpty()) {
            LOGGER.info("No submission found for submission id {}", listOneSubmissionCommand.submissionId);
        } else {
            writeData(submission.get(), getTargetFolderPath(listOneSubmissionCommand));
        }
    }

    private void listSubmissions(final ListAllSubmissionsCommand listAllSubmissionsCommand) {
        final var destinationId = listAllSubmissionsCommand.destinationId;
        LOGGER.info("Listing available submissions for destination {}", destinationId);
        for (final SubmissionForPickup submission : subscriberClient.getAvailableSubmissions(destinationId)) {
            LOGGER.info("caseId: {} - submissionId: {}", submission.getCaseId(), submission.getSubmissionId());
        }
    }

    private Optional<SubmitSubmission> sendSubmission(final SendSubmissionCommand sendSubmissionCommand) throws IOException {
        LOGGER.info("Sending new submission to destination {}", sendSubmissionCommand.destinationId);
        final List<File> attachments = sendSubmissionCommand.attachments.stream().map(File::new).collect(Collectors.toList());
        if (sendSubmissionCommand.dataType == AttachmentDataType.json) {
            return sendWithJsonData(sendSubmissionCommand, attachments);
        } else {
            return sendWithXmlData(sendSubmissionCommand, attachments);
        }
    }

    private void sendBatch(final SendBatchCommand sendBatchCommand) throws BatchImportException, IOException {
        final List<ImportRecord> importRecords = batchImporter.readRecords(sendBatchCommand.dataPath);
        LOGGER.info("Sending batch of {} submissions", importRecords.size());
        var importCount = 0;
        for (final ImportRecord importRecord : importRecords) {
            if (sendSubmission(mapToSubmissionToSend(importRecord)).isPresent()) {
                LOGGER.info(IMPORTED_RECORD_TEMPLATE, importCount + 1);
                importCount++;
            } else {
                LOGGER.info("Failed importing record {} => {}", importCount + 1, importRecord);
            }
        }
        LOGGER.info("DONE ! Finished batch import of {} submissions", importCount);
    }

    private Optional<SubmitSubmission> sendWithJsonData(final SendSubmissionCommand sendSubmissionCommand, final List<File> attachments) throws IOException {
        return senderClient.newSubmission()
                .withAttachments(attachments)
                .withJsonData(getDataAsString(sendSubmissionCommand.data))
                .withDestination(sendSubmissionCommand.destinationId)
                .withServiceType(sendSubmissionCommand.serviceName, sendSubmissionCommand.leikaKey)
                .submit();
    }

    private Optional<SubmitSubmission> sendWithXmlData(final SendSubmissionCommand sendSubmissionCommand, final List<File> attachments) throws IOException {
        return senderClient.newSubmission()
                .withAttachments(attachments)
                .withXmlData(getDataAsString(sendSubmissionCommand.data))
                .withDestination(sendSubmissionCommand.destinationId)
                .withServiceType(sendSubmissionCommand.serviceName, sendSubmissionCommand.leikaKey)
                .submit();
    }

    private String getTargetFolderPath(final ListOneSubmissionCommand listOneSubmissionCommand) {
        if (listOneSubmissionCommand.targetFolder != null) {
            return listOneSubmissionCommand.targetFolder + "/" + listOneSubmissionCommand.submissionId;
        } else {
            return listOneSubmissionCommand.submissionId.toString();
        }
    }

    private String getDataAsString(final String dataPath) throws IOException {
        LOGGER.info("Reading data from {}", dataPath);
        return Files.readString(Path.of(dataPath));
    }

    private void writeData(final ReceivedSubmission submissionData, final String dataDirPath) throws IOException {
        LOGGER.info("Creating data directory for submission in {}", dataDirPath);
        Files.createDirectories(Path.of(dataDirPath));

        final var fileEnding = AttachmentDataType.getFileTypeFromMimeType(submissionData.getMimeType());
        final var filePath = Path.of(dataDirPath + "/data." + fileEnding);
        LOGGER.info("Writing data.{}", fileEnding);
        Files.write(filePath, submissionData.getData().getBytes(StandardCharsets.UTF_8));

        for (final AttachmentWithData attachment : submissionData.getAttachments()) {
            final String filename = attachment.getAttachmentMetadata().getFilename();
            LOGGER.info("Writing attachment {}", filename);
            Files.write(Path.of(dataDirPath, filename), attachment.getDecryptedData());
        }
    }

    private SendSubmissionCommand mapToSubmissionToSend(final ImportRecord importRecord) {
        final var submission = new SendSubmissionCommand();
        submission.setDestinationId(importRecord.getDestinationId());
        submission.setServiceName(importRecord.getServiceName());
        submission.setLeikaKey(importRecord.getLeikaKey());
        submission.setData(importRecord.getData().trim());
        submission.setDataType(importRecord.getDataType());
        submission.setAttachments(Stream.of(importRecord.getAttachments().split(",")).map(String::trim).collect(Collectors.toList()));
        return submission;
    }
}
